/*
 * Copyright (c) 1999-2021 Hartmut Seichter
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#ifndef PW_CORE_MATRIXBASE_HPP
#define PW_CORE_MATRIXBASE_HPP

#include <algorithm>
#include <iterator>
#include <numeric>
#include <type_traits>
#include <cmath>

namespace pw {

template <typename T, typename Derived>
struct matrixbase_ {

    using value_type = T;
    using derived_type = Derived;
    using iterator = value_type*;
    using const_iterator = const value_type*;

	constexpr Derived& derived() noexcept { return static_cast<Derived&>(*this); }
    constexpr const Derived& derived() const noexcept { return static_cast<const Derived&>(*this); }

    constexpr std::size_t size() const noexcept {
        return std::extent<decltype(Derived::data)>::value;
    }

    constexpr Derived& fill(const T& v) noexcept {
        std::fill(std::begin(derived().data), std::end(derived().data), T(v));
        return derived();
    }

    constexpr Derived& zero() {
        return derived().fill(0);
    }

	constexpr T squared_norm() const noexcept {
		return std::accumulate(std::begin(derived().data),std::end(derived().data), T(0),
							   [&](const T& a,const T& b){
			return a + b * b;
		});
	}

    static constexpr Derived
    all(const std::convertible_to<T> auto& v) noexcept {
        Derived d;
        d.fill(v);
        return d;
    }

	constexpr T norm() const noexcept {
		return std::sqrt(squared_norm());
	}

    constexpr Derived normalized() const noexcept {
		return *this / this->norm() ;
	}

	constexpr void normalize() {
		*this /= this->norm();
	}


    constexpr iterator begin() { return &derived().data[0]; }
    constexpr iterator end() { return &derived().data[0] + size(); }
    constexpr const_iterator begin() const { return &derived().data[0]; }
    constexpr const_iterator end() const { return &derived().data[0] + size(); }

    constexpr T& operator [] (std::size_t i) {
        return derived().data[i];
    }

    constexpr const T& operator [] (std::size_t i) const {
        return derived().data[i];
    }

    static constexpr T dot(const Derived &a,const Derived &b) noexcept
    {
        return std::inner_product(std::begin(a),std::end(a),std::begin(b),T(0));
	}

    static constexpr Derived lerp(const Derived &a,const Derived &b,const T& t) noexcept
    {
		return a + (b - a) * t;
	}


    //
    // M (op) scalar 
    //
    constexpr Derived operator*(const T& b) const {
        Derived r(derived());
        for (auto& e : r)
            e *= b;
        return r;
    }
    constexpr Derived operator/(const T& b) const {
        Derived r(derived());
        for (auto& e : r)
            e /= b;
        return r;
    }
    constexpr Derived operator+(const T& b) const {
        Derived r(derived());
        for (auto& e : r)
            e += b;
        return r;
    }
    constexpr Derived operator-(const T& b) const {
        Derived r(derived());
        for (auto& e : r)
            e -= b;
        return r;
    }

    //
    // M (op)== N
    //

    constexpr void operator*=(const std::convertible_to<T> auto& b) {
        for (auto& e : derived())
            e *= b;
    }
    constexpr void operator/=(const std::convertible_to<T> auto& b) {
        for (auto& e : derived())
            e /= b;
    }
    constexpr void operator+=(const std::convertible_to<T> auto& b) {
        for (auto& e : derived())
            e += b;
    }
    constexpr void operator-=(const std::convertible_to<T> auto& b) {
        for (auto& e : derived())
            e -= b;
    }

    //
    // M (op) N
    //

    /// @brief in-place addition with matrix of same size
    /// @param rhs righthand-side matrix
    constexpr void operator+=(const Derived& b) noexcept {
        std::transform(b.cbegin(), b.cend(), (*this).cbegin(), (*this).begin(),
                       std::plus<>{});
    }

    /// @brief in-place subtraction with matrix of same size
    /// @param rhs righthand-side matrix
    constexpr void operator-=(const Derived& b) noexcept {
        std::transform(b.cbegin(), b.cend(), (*this).cbegin(), (*this).begin(),
                       std::minus<>{});
    }

    /// @brief addition with matrix of same size
    /// @param rhs righthand-side matrix
    constexpr Derived operator+(const std::convertible_to<Derived> auto& rhs) const noexcept {
        Derived d{derived()};
        std::transform(std::begin(rhs), std::end(rhs), (d).begin(), std::plus<>{});
        return d;
    }

    /// @brief subtraction with matrix of same size
    /// @param rhs righthand-side matrix
    constexpr Derived operator-(const Derived& rhs) const noexcept {
        Derived d(derived());
        d -= rhs;
        return d;
    }


};

}


#endif
